/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2011, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */


#include "cc_mgr.h"
#include <time.h>

#ifdef PROD_TREE_ROOT
#include <md_mod_reg.h>
#endif

#define LOCK_NOT_HELD      0
#define LOCK_HELD_BY_MGR   1
#define LOCK_HELD_BY_STAT  2

#define MAX_DISABLE_ERROR_COUNT 4

#define IS_ConnectX_4(device_id) (device_id == 0x1013 ? true : false)
#define IS_ConnectX_5(device_id) (device_id == 0x1017 ? true : false)
#define IS_ConnectX_6(device_id) (device_id == 0x101b ? true : false)
#define IS_BF1(device_id) (device_id == 0xa2d2 ? true : false)
#define IS_Eagle(device_id) (device_id == 0xcf08 ? true : false)
#define IS_Quantum(device_id) (device_id == 0xd2f0 ? true : false)

/****************************************************/
extern "C" {
    /*
     * Wrapper to pass statistics and other data to smd plugin
     */
    void do_smd_report(uint32_t msg_type, const int cnt, void  *msg_data);
}


/****************************************************/
/**************** private methods *******************/
/****************************************************/
//============================================================================//
void CongestionControlManager::MgrLockCCMutex()
{
    pthread_mutex_lock(&m_lock_holder_mutex);
    if(m_lock_holder == LOCK_NOT_HELD) {
    pthread_mutex_lock(&m_cc_mutex);
    m_lock_holder = LOCK_HELD_BY_MGR;
    }
    else if (m_lock_holder == LOCK_HELD_BY_STAT) {
    pthread_mutex_unlock(&m_lock_holder_mutex);
    pthread_mutex_lock(&m_cc_gate_mutex);
    m_cc_mgr_request = true;
    pthread_mutex_lock(&m_cc_mutex);
    pthread_mutex_unlock(&m_cc_gate_mutex);
    pthread_mutex_lock(&m_lock_holder_mutex);
    m_lock_holder = LOCK_HELD_BY_MGR;
    }
    // else - manager is already holding the lock
    pthread_mutex_unlock(&m_lock_holder_mutex);
}


void CongestionControlManager::UnLockCCMutex(uint8_t lock_holder)
{
    pthread_mutex_lock(&m_lock_holder_mutex);
    //
    // Unlock only if it is really locked and the current
    // thread is the one that is holding the lock.
    //
    if (m_lock_holder == lock_holder) {
        pthread_mutex_unlock(&m_cc_mutex);
        m_lock_holder = LOCK_NOT_HELD;
    }
    pthread_mutex_unlock(&m_lock_holder_mutex);
}


void CongestionControlManager::StatLockCCMutex()
{
    pthread_mutex_lock(&m_cc_mutex);
    pthread_mutex_lock(&m_lock_holder_mutex);
    m_lock_holder = LOCK_HELD_BY_STAT;
    pthread_mutex_unlock(&m_lock_holder_mutex);
}


//============================================================================//
void CongestionControlManager::CheckRC(int& rc)
{
    struct timeval curr_time;

    //Do not stop disable process in the middle
    if (m_enable == false) {
        return;
    }
    // Return if error checking mechanism is disabled
    if (!m_error_window)
        return;

    if (rc == IBIS_MAD_STATUS_SUCCESS)
        return;

    //
    // The error window is supposed to prevent the CC Manager
    // from stealing the CPU time from the SM and wasting it
    // on errors/timeouts.
    // We will NOT treat errors in the response MAD as a
    // problem in the CC Manager execution cylcle and will
    // not count them in the error window.
    //
    if ((rc != IBIS_MAD_STATUS_SEND_FAILED) &&
        (rc != IBIS_MAD_STATUS_RECV_FAILED) &&
        (rc != IBIS_MAD_STATUS_TIMEOUT))
        return;

    m_num_errors++;
    if (!m_max_errors) {
        // Abort on first error
        goto check_rc_error;
    }
    gettimeofday(&curr_time, NULL);

    m_oldest_error = (m_oldest_error + 1) % m_max_errors;
    if (m_p_error_window[m_oldest_error].tv_sec && // If equals to zero,
                                                   // no error happened, thus
                                                   // not enough errors happened
        curr_time.tv_sec - m_p_error_window[m_oldest_error].tv_sec <= m_error_window) {
        goto check_rc_error;
    }
    // Override oldest error with the new error
    m_p_error_window[m_oldest_error] = curr_time;

    return;

check_rc_error:
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
            "Exited because reached %d error's in less than %ld seconds\n",
            m_num_errors, m_error_window);
    throw 1;
}


//============================================================================//
int CongestionControlManager::ParseConfFileName(char* osm_plugin_options)
{
    int rc = 0;
    if (osm_plugin_options == NULL) {
        return 0;
    }

    string osm_plugin_options_str(osm_plugin_options);
    string conf_file_flag(CONF_INPUT_FILE_FLAG);
    string::size_type flag_pos = 0, start_pos = 0, last_pos = 0;

    if (osm_plugin_options_str.empty()) {
        return 0;
    }
    if ((flag_pos =
         osm_plugin_options_str.find(conf_file_flag)) == string::npos) {
        return 1;
    }
    if ((start_pos =
         osm_plugin_options_str.find_first_not_of(" ", flag_pos + conf_file_flag.length()))
         == string::npos) {
        return 1;
    }
    if ((last_pos =
         osm_plugin_options_str.find_first_of(" ", start_pos)) == string::npos) {
        last_pos = osm_plugin_options_str.length();
    }
    last_pos--;

    m_conf_file_name = osm_plugin_options_str.substr(start_pos, last_pos - start_pos + 1);
    return rc;
}


void CongestionControlManager::InitUserOptions()
{
    // General Options
    m_general_user_options["enable"]              = ENABLE;
    m_general_user_options["cc_key"]              = CC_KEY;
    m_general_user_options["num_hosts"]           = NUM_HOSTS;
    m_general_user_options["max_errors"]          = MAX_ERRORS;
    m_general_user_options["error_window"]        = ERROR_WINDOW;
    m_general_user_options["cc_statistics_cycle"] = CC_STATISTICS_CYCLE;
    m_general_user_options["cc_log_file_name"]    = CC_LOG_FILE_NAME;
    m_general_user_options["cc_log_file_size"]    = CC_LOG_FILE_SIZE;

    // SW User Options
    m_sw_user_options["enable"]       = SW_ENABLE;
    m_sw_user_options["threshold"]    = THRESHOLD;
    m_sw_user_options["packet_size"]  = PACKET_SIZE;
    m_sw_user_options["marking_rate"] = MARKING_RATE;

    // CA User Options
    m_ca_user_options["enable"]            = CA_ENABLE;
    m_ca_user_options["port_control"]      = PORT_CONTROL;
    m_ca_user_options["ca_control_map"]    = CA_CONTROL_MAP;
    m_ca_user_options["ccti_timer"]        = CCTI_TIMER;
    m_ca_user_options["ccti_increase"]     = CCTI_INCR;
    m_ca_user_options["trigger_threshold"] = TRIGGER_THRESHOLD;
    m_ca_user_options["ccti_min"]          = CCTI_MIN;
    m_ca_user_options["cct"]               = CCT;
}


void CongestionControlManager::InitMasterDataBase()
{
    m_master_db.m_enable              = CC_GENERAL_DEFAULT_ENABLE;
    m_master_db.m_cc_key              = CC_GENERAL_DEFAULT_CC_KEY;
    m_master_db.m_error_window        = CC_GENERAL_DEFAULT_ERROR_WINDOW;
    m_master_db.m_max_errors          = CC_GENERAL_DEFAULT_MAX_ERRORS;
    m_master_db.m_cc_statistics_cycle = CC_GENERAL_DEFAULT_CC_STATISTICS_CYCLE;
    m_master_db.m_cc_log_file_name    = CC_GENERAL_DEFAULT_CC_LOG_FILE_NAME;
    m_master_db.m_cc_log_file_size    = CC_GENERAL_DEFAULT_CC_LOG_FILE_SIZE;
    m_master_db.m_num_hosts           = 0;
    m_num_errors		      = 0;
    m_oldest_error		      = 0;

    InitCC_SwitchCongestionSetting(m_master_db.m_sw_default_params);
    InitCC_CACongestionSetting(m_master_db.m_ca_default_params);
}


void CongestionControlManager::Construct(char *osm_plugin_options)
{
    ParseConfFileName(osm_plugin_options);
    OSM_CC_LOG(this->m_p_osm_log, OSM_LOG_INFO, "Configuration file is %s\n",
            this->m_conf_file_name.c_str());
    fprintf(stdout, "CC Manager - Configuration file is %s\n",
            this->m_conf_file_name.c_str());
    InitUserOptions();
    InitMasterDataBase();
    pthread_mutex_init(&m_cc_mutex, NULL);
    pthread_mutex_init(&m_cc_gate_mutex, NULL);
    pthread_mutex_init(&m_lock_holder_mutex, NULL);
}


//============================================================================//
bool CongestionControlManager::CheckNodeType(const u_int64_t port_guid,
        u_int8_t& node_type)
{
    osm_port_t *p_osm_port;
    if ((p_osm_port = osm_get_port_by_guid(m_p_osm_subn,cl_hton64(port_guid)))) {
        node_type = osm_node_get_type(p_osm_port->p_node);
        return true;
    }
    return false;
}


void CongestionControlManager::RemoveAbsentNodes()
{
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator aux_ca_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator aux_sw_node_it;
    u_int8_t node_type;

    sw_node_it = m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=  m_sw_cc_setting_db.m_nodes_db.end()) {
        if (CheckNodeType(sw_node_it->first,node_type) && node_type == IB_NODE_TYPE_SWITCH) {
            sw_node_it++;
            continue;
        }
        //Remove non existing SW nodes
        aux_sw_node_it = sw_node_it;
        sw_node_it++;
        m_sw_cc_setting_db.m_nodes_db.erase(aux_sw_node_it);
    }

    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()) {
        if (CheckNodeType(ca_node_it->first,node_type) && node_type == IB_NODE_TYPE_CA) {
            ca_node_it++;
            continue;
        }
        //Remove none existing CA nodes
        aux_ca_node_it = ca_node_it;
        ca_node_it++;
        m_ca_cc_setting_db.m_nodes_db.erase(aux_ca_node_it);
    }
}


//============================================================================//
int CongestionControlManager::Init()
{
    static bool already_initialized = false;

    if (already_initialized)
        return 0;

    //open log file
    if (m_cc_log_file_size) {
        // Only if the file size is greater than 0,
        // the log file will be created.
        if (tt_log_construct_v2(0xff,           // need_to_flush
                m_cc_log_file_name.c_str(),     // log file
                m_cc_log_file_size,             // max size (in MB)
                0)) {                           // append to prev log
            m_cc_log_file_size = 0;
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Error opening Congestion Control log file : %s\n", m_cc_log_file_name.c_str());
            throw 1;
        }

        tt_log_set_level_verbosity(tt_log_get_level_verbosity() |
                                   TT_LOG_LEVEL_ERROR | TT_LOG_LEVEL_INFO);

        OSM_CC_LOG(m_p_osm_log, OSM_LOG_INFO, "Congestion Control log file: %s\n",
                m_cc_log_file_name.c_str());
    }

    //init ibis
    if (true == m_ibis_obj.IsInit()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG, "ibis already initialized\n");
    }
    else if (m_ibis_obj.Init()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR, "ibis initialization failed\n");
        throw 1;
    }
    else {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "ibis initialized\n");
    }

    //set port
    m_port_guid = cl_ntoh64(m_p_osm_vendor->umad_port.port_guid);
    if (true == m_ibis_obj.IsReady()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG, "ibis SetPort ready\n");
    }
    else if (m_ibis_obj.SetPort(ADB2C_CPU_TO_BE64(m_port_guid))) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR, "ibis SetPort failed\n");
        throw 1;
    }
    else {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "ibis running on GUID " U64H_FMT"\n", m_port_guid);
    }

    already_initialized = true;
    return 0;
}


//============================================================================//
bool CongestionControlManager::IsDeviceIDSupported(const CCNodeInfo& node_info)
{
    u_int16_t device_id = cl_ntoh16(node_info.m_p_osm_node->node_info.device_id);
    bool is_supported = m_ibis_obj.IsDevShaldag(device_id) ||
                        m_ibis_obj.IsDevSwitchXIB(device_id) ||
                        m_ibis_obj.IsDevConnectXIB(device_id) ||
                        m_ibis_obj.IsDevPelican(device_id) ||
                        m_ibis_obj.IsDevGolan(device_id) ||
                        IS_ConnectX_4(device_id) ||
                        IS_ConnectX_5(device_id) ||
                        IS_ConnectX_6(device_id) ||
                        IS_BF1(device_id) ||
                        IS_Eagle(device_id) ||
                        IS_Quantum(device_id);

    if (!is_supported) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Node " U64H_FMT " : device ID %u is not supported\n",
                node_info.m_port_guid, device_id);
    //
    // Note that although CC is not supported by this node,
    // in case it's an HCA, we are still counting it as a source
    // of traffic for the num_hosts (unless this number was
    // provided by a user, in which case no need to count anyway).
    //
    }
    return is_supported;
}


int CongestionControlManager::GetNodeOSMUpdateNeeded(const CCNodeInfo& node_info,
        bool& osm_update_needed)
{
    uint8_t node_type = osm_node_get_type(node_info.m_p_osm_node);
    if ((node_type == IB_NODE_TYPE_CA) || (node_type == IB_NODE_TYPE_SWITCH)) {
        osm_update_needed = node_info.m_p_osm_physp->need_update ||
                            m_p_osm_subn->need_update;
        return 0;
    }
    return 1;
}


int CongestionControlManager::GetNodeMTUSize(const CCNodeInfo& node_info,
        u_int32_t& mtu_size)
{
    int rc = 0;
    mtu_size = strtoul(osm_get_mtu_str(
        ib_port_info_get_neighbor_mtu(&node_info.m_p_osm_physp->port_info)),0,0);
    if (!mtu_size) {
        // if strtoul failed assign the default value
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Failed to get mtu_size for Node with GUID " U64H_FMT ". Setting default MTU size: %u\n",
                node_info.m_port_guid,
                CC_GENERAL_DEFAULT_MTU_SIZE);

        mtu_size = CC_GENERAL_DEFAULT_MTU_SIZE;
    }
    return rc;
}


int CongestionControlManager::UpdateCANode(const CCNodeInfo& node_info)
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_ca_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
                   m_ca_cc_setting_db.m_nodes_db.end()) {
        if (!IsDeviceIDSupported(node_info)) {
            return rc;
        }
        CACCSettingDataBaseNodeEntry new_ca_node(node_info.m_port_guid,
                                                 m_ca_cc_setting_db.m_default_params);
        m_ca_cc_setting_db.m_nodes_db.insert(
            pair<u_int64_t, CACCSettingDataBaseNodeEntry>(node_info.m_port_guid, new_ca_node));
        node_it = m_ca_cc_setting_db.m_nodes_db.find(node_info.m_port_guid);
    }
    if (!IsDeviceIDSupported(node_info)) {
        m_ca_cc_setting_db.m_nodes_db.erase(node_it);
        return rc;
    }
    node_it->second.m_node_info = node_info;
    if (m_osm_event_id == OSM_EVENT_ID_HEAVY_SWEEP_DONE) {
        // Update osm_update_needed field on this event
        GetNodeOSMUpdateNeeded(node_info, node_it->second.m_osm_update_needed);
    }
    GetNodeMTUSize(node_info, node_it->second.m_mtu_size);
    return rc;
}


int CongestionControlManager::UpdateSWNode(const CCNodeInfo& node_info)
{
    int rc = 0;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_sw_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
                   m_sw_cc_setting_db.m_nodes_db.end()) {
        if (!IsDeviceIDSupported(node_info)) {
            return rc;
        }
        SWCCSettingDataBaseNodeEntry new_sw_node(node_info.m_port_guid,
                                                 m_sw_cc_setting_db.m_default_params);
        m_sw_cc_setting_db.m_nodes_db.insert(
            pair<u_int64_t, SWCCSettingDataBaseNodeEntry>(node_info.m_port_guid,new_sw_node));
        node_it = m_sw_cc_setting_db.m_nodes_db.find(node_info.m_port_guid);
    }
    if (!IsDeviceIDSupported(node_info)) {
        m_sw_cc_setting_db.m_nodes_db.erase(node_it);
        return rc;
    }
    node_it->second.m_node_info = node_info;
    if (m_osm_event_id == OSM_EVENT_ID_HEAVY_SWEEP_DONE) {
        // Update osm_update_needed field on this event
        GetNodeOSMUpdateNeeded(node_info, node_it->second.m_osm_update_needed);
    }
    return rc;
}


uint8_t CongestionControlManager::GetSL(osm_port_t *p_dest_port)
{
      osm_path_parms_t path_parms;
      ib_api_status_t status;

      memset(&path_parms, 0, sizeof(osm_path_parms_t));

      status = osm_get_path_params(m_p_osm_sa,
            osm_get_port_by_guid(m_p_osm_subn, m_p_osm_subn->sm_port_guid),// src port (always SM port)
            cl_ntoh16(m_p_osm_subn->sm_base_lid),
            p_dest_port,
            cl_ntoh16(osm_port_get_base_lid(p_dest_port)),
            &path_parms);

      if (!status) {
          return path_parms.sl;
      } else {
          return m_p_osm_subn->opt.sm_sl;
      }
}


//============================================================================//
int CongestionControlManager::CheckNodeSupportsCCoFabric(const CCNodeInfo& node_info,
        bool& node_supports_cc)
{
    int rc = 0;
    struct IB_ClassPortInfo ib_class_port_info;
    if ((rc = m_ibis_obj.CCClassPortInfoGet(node_info.m_lid,
                                            node_info.m_sl,
                                            m_cc_key,
                                            &ib_class_port_info))) {
        //
        // Treating failure in getting CC ClassPortInfo as a case
        // of node not supporting CC and not as error per say.
        //
        CheckRC(rc);
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Error getting CC ClassPortInfo (node GUID " U64H_FMT
                ") - assuming no CC support\n",
                node_info.m_port_guid);
        node_supports_cc = false;
        return rc;
    }
    if (ib_class_port_info.BaseVersion != 1 || ib_class_port_info.ClassVersion != 2 ) {
        node_supports_cc = false;
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Node with GUID " U64H_FMT " doesn't support Congestion Control\n",
                node_info.m_port_guid);
        return rc;
    }
    node_supports_cc = true;
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
            "Node with GUID " U64H_FMT " supports Congestion Control\n",
            node_info.m_port_guid);
    return rc;
}


int CongestionControlManager::GetCANodeCCSupport(const CCNodeInfo& node_info,
        bool& node_supports_cc)
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_ca_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
        m_ca_cc_setting_db.m_nodes_db.end()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to find CA node with GUID:" U64H_FMT "\n",
                node_info.m_port_guid);
        return 1;
    }
    node_it->second.m_node_info = node_info;
    if (node_it->second.m_cc_support == CC_SUPPORT_UNKNOWN) {
        rc = CheckNodeSupportsCCoFabric(node_info, node_supports_cc);
        if (rc) {
            return rc;
        }
        node_it->second.m_cc_support =
        (node_supports_cc) ? CC_SUPPORTED : CC_NOT_SUPPORTED;
    }
    else {
        node_supports_cc =
        (node_it->second.m_cc_support == CC_SUPPORTED)
        ? true : false;
    }
    return rc;
}


int CongestionControlManager::GetSWNodeCCSupport(const CCNodeInfo& node_info,
        bool& node_supports_cc)
{
    int rc = 0;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_sw_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
                   m_sw_cc_setting_db.m_nodes_db.end()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to find Switch node with GUID:" U64H_FMT "\n",
                node_info.m_port_guid);
        return 1;
    }
    if (node_it->second.m_cc_support == CC_SUPPORT_UNKNOWN) {
        rc = CheckNodeSupportsCCoFabric(node_info, node_supports_cc);
        if (rc) {
            return rc;
        }
        node_it->second.m_cc_support =
        (node_supports_cc) ? CC_SUPPORTED : CC_NOT_SUPPORTED;
    }
    else {
        node_supports_cc =
        (node_it->second.m_cc_support == CC_SUPPORTED)
        ? true : false;
    }
    return rc;
}


int CongestionControlManager::CollectCCSupportForAllNodes()
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    bool node_cc_support;

    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Collect CC Support For %u switches and %u HCAs\n",
            (int)m_sw_cc_setting_db.m_nodes_db.size(),
            (int)m_ca_cc_setting_db.m_nodes_db.size());

    sw_node_it = m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=  m_sw_cc_setting_db.m_nodes_db.end()) {
        GetSWNodeCCSupport(sw_node_it->second.m_node_info,node_cc_support);
        sw_node_it++;
    }

    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()) {
        GetCANodeCCSupport(ca_node_it->second.m_node_info,node_cc_support);
        ca_node_it++;
    }
    return rc;
}


//============================================================================//
/***************/
int CongestionControlManager::GetCAControlTableCap(const CCNodeInfo& node_info,
        unsigned int& control_table_cap)
{
    int rc = 0;
    struct CC_CongestionInfo cc_congestion_info;
    if (!control_table_cap) {
        // If zero, wasn't user defined
        if ((rc = m_ibis_obj.CCCongestionInfoGet(node_info.m_lid,
                                                 node_info.m_sl,
                                                 m_cc_key,
                                                 &cc_congestion_info))) {
            CheckRC(rc);
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Failed to send CongestionInfo [Get] to node GUID:"
                    U64H_FMT "\n", node_info.m_port_guid);
            return rc;
        }
        control_table_cap = cc_congestion_info.ControlTableCap;
    }
    return rc;
}

void CongestionControlManager::GetCAControlTableCapForAllNodes()
{
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()){
        if (ca_node_it->second.m_cc_support == CC_SUPPORTED &&
            ca_node_it->second.m_node_ccti_list.size() == 0 &&
            ( m_enable || ca_node_it->second.m_error_count < MAX_DISABLE_ERROR_COUNT)) {
            GetCAControlTableCap(ca_node_it->second.m_node_info,
                                 ca_node_it->second.m_control_table_cap);
        }
        ca_node_it++;
    }
    return;
}


/***************/
void CongestionControlManager::DumpCACongSetting(struct CC_CACongestionSetting& cc_ca_congestion_setting)
{
    char buff[128];
    stringstream output_ss;
    for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
        sprintf(buff,"[sl:%2d] Trigger_Threshold: 0x%02x, CCTI_Min: %3u, CCTI_Increase: %3u,"
                     " CCTI_Timer: 0x%04x.",
                sl,
                cc_ca_congestion_setting.CACongestionEntryList.CACongestionEntryListElement[sl].Trigger_Threshold,
                cc_ca_congestion_setting.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Min,
                cc_ca_congestion_setting.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Increase,
                cc_ca_congestion_setting.CACongestionEntryList.CACongestionEntryListElement[sl].CCTI_Timer);
        output_ss << buff << PRINT_PREFIX;
    }
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "\n" PRINT_PREFIX
            "Control_Map: " U32H_FMT ", Port_Control:" U32H_FMT PRINT_PREFIX
            "%s\n",
            cc_ca_congestion_setting.Control_Map,
            cc_ca_congestion_setting.Port_Control,
            output_ss.str().c_str());
}

int CongestionControlManager::CalculateCCTITimer(const CCNodeInfo& node_info,
        const u_int32_t mtu_size,
        u_int16_t& ccti_timer)
{
    int rc = 0;
    u_int32_t mtuTime_us = mtu_size / 1000;
    u_int32_t decTime_us = 2 * mtuTime_us * m_num_hosts;
    double timer = (decTime_us + 1.023) / 1.024;
    ccti_timer = (u_int16_t) timer ;
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
            "Calculated timer :" U32H_FMT " for Node with GUID " U64H_FMT "\n",
            ccti_timer,
            node_info.m_port_guid);

    return rc;
}

int CongestionControlManager::GetCANodeCCSettings(const CCNodeInfo& node_info,
        bool& node_supports_cc,
        struct CC_CACongestionSetting& node_params,
        map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator& ca_node_it)
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_ca_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
        m_ca_cc_setting_db.m_nodes_db.end()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to find CA node with GUID:" U64H_FMT "\n",
                node_info.m_port_guid);
        return 1;
    }
    ca_node_it = node_it;
    node_supports_cc = (node_it->second.m_cc_support == CC_SUPPORTED);
    if (!node_supports_cc) {
        return rc;
    }
    if (node_it->second.m_is_set_flags.find("ccti_timer") ==
        node_it->second.m_is_set_flags.end() &&
        node_it->second.m_enable) {
        // If ccti_timer wasn't configured by user, need to calculate it
        rc = CalculateCCTITimer(node_info,
                                node_it->second.m_mtu_size,
                                node_it->second.m_node_params.
                                CACongestionEntryList.CACongestionEntryListElement[0].CCTI_Timer);
    }
    for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
        node_it->second.m_node_params.CACongestionEntryList.
            CACongestionEntryListElement[sl].CCTI_Timer =
        node_it->second.m_node_params.CACongestionEntryList.
            CACongestionEntryListElement[0].CCTI_Timer;
    }
    node_params = node_it->second.m_node_params;
    return rc;
}

bool CongestionControlManager::CompareCACongestionSetting(const struct CC_CACongestionSetting &lhs,
        const struct CC_CACongestionSetting &rhs)
{
    if (lhs.Control_Map != rhs.Control_Map) {
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "CA Control_Map mismatch. %d : %d\n",
                lhs.Control_Map,
                rhs.Control_Map);
            return false;
    }
    if (lhs.Port_Control != rhs.Port_Control) {
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                "CA Port_Control mismatch. %d : %d\n",
                lhs.Port_Control,
                rhs.Port_Control);
            return false;
    }
    for (unsigned int sl = 0; sl < SL_MAX_NUM; sl++) {
        if (lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            CCTI_Increase !=
            rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            CCTI_Increase) {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                    "CA CCTI_Increase mismatch on sl %d. %d : %d\n",
                    sl,
                    lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    CCTI_Increase,
                    rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    CCTI_Increase);
                return false;
        }
        if (lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            CCTI_Min !=
            rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            CCTI_Min) {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                    "CA CCTI_Min mismatch on sl %d. %d : %d\n",
                    sl,
                    lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    CCTI_Min,
                    rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    CCTI_Min);
                return false;
        }
        if (lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            CCTI_Timer !=
            rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            CCTI_Timer) {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                    "CA CCTI_Timer mismatch on sl %d. %d : %d\n",
                    sl,
                    lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    CCTI_Timer,
                    rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    CCTI_Timer);
                return false;
        }
        if (lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            Trigger_Threshold !=
            rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
            Trigger_Threshold) {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                    "CA Trigger_Threshold mismatch on sl %d. %d : %d\n",
                    sl,
                    lhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    Trigger_Threshold,
                    rhs.CACongestionEntryList.CACongestionEntryListElement[sl].
                    Trigger_Threshold);
                return false;
        }
    }
    return true;
}

int CongestionControlManager::SetCACongestionSetting(const CCNodeInfo& node_info)
{
    int rc = 0;
    struct CC_CACongestionSetting cc_ca_congestion_setting;
    bool node_supports_cc = false;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    rc = GetCANodeCCSettings(node_info, node_supports_cc, cc_ca_congestion_setting, ca_node_it);
    if (rc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Node with GUID " U64H_FMT " can't be configured. "
                "Congestion Control information can't be retrieved from "
                "the node\n", node_info.m_port_guid);
        return rc;
    }
    if (!node_supports_cc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Node with GUID " U64H_FMT " doesn't support Congestion "
                "Control\n", node_info.m_port_guid);
        return rc;
    }
    rc = m_ibis_obj.CCCACongestionSettingSet(node_info.m_lid,
                                             node_info.m_sl,
                                             m_cc_key,
                                             &cc_ca_congestion_setting);
    CheckRC(rc);
    if (!rc) {
        ca_node_it->second.m_is_congestion_settings_set = true;
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Set CA node with GUID " U64H_FMT
                " with the following CACongestionSetting parameters:\n", node_info.m_port_guid);
        DumpCACongSetting(cc_ca_congestion_setting);
    }
    else {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to send CACongestionSetting [Set] to node GUID:"
                U64H_FMT "\n", node_info.m_port_guid);
        if (m_enable == false) {
            ca_node_it->second.m_error_count++;
        }
    }
    return rc;
}

void CongestionControlManager::SetCACongestionSettingForAllNodes()
{
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()){
        if (ca_node_it->second.m_cc_support == CC_SUPPORTED) {
            if (!ca_node_it->second.m_is_congestion_settings_set ||
                ca_node_it->second.m_osm_update_needed) {

                if (m_enable || ca_node_it->second.m_error_count < MAX_DISABLE_ERROR_COUNT) {                
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                            "CA node with GUID " U64H_FMT ". "
                            PRINT_PREFIX "m_osm_update_needed: %d, m_is_congestion_settings_set: %d\n",
                            ca_node_it->second.m_node_info.m_port_guid,
                            ca_node_it->second.m_osm_update_needed,
                            ca_node_it->second.m_is_congestion_settings_set);
    
                    SetCACongestionSetting(ca_node_it->second.m_node_info);
                } else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                            "Do not set disable CACongestionSetting for CA node with GUID " U64H_FMT
                            "because of previous error.\n", ca_node_it->second.m_node_info.m_port_guid);
                }
            } else {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                        "No need to set CACongestionSetting for CA node with GUID " U64H_FMT
                        "\n", ca_node_it->second.m_node_info.m_port_guid);
            }
        }
        ca_node_it++;
    }
    return;
}


/***************/
void CongestionControlManager::DumpCCTableSetting(struct CC_CongestionControlTable& cc_congestion_control_table)
{
    u_int16_t value;
    stringstream ss;
    stringstream out_ss;
    for (unsigned int idx = 0; idx < ccti_entry_list_elements_num; idx++) {
        value =
        (((cc_congestion_control_table.CCTI_Entry_List.CCTI_Entry_ListElement[idx].CCT_Shift) & 0x3) << 14) |
        (cc_congestion_control_table.CCTI_Entry_List.CCTI_Entry_ListElement[idx].CCT_Multiplier & 0x3fff);
        ss << setfill('0') << "[" << setw(2) << std::dec << idx << setw(0);
        ss << "] 0x" << setw(4) << std::hex << value << setw(0) << ", ";
        if (idx && ((idx % 8) == 7 /*|| (idx % 8) == 3*/)) {
            out_ss << ss.str() << PRINT_PREFIX;
            ss.str("");
        }
    }

    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "\n" PRINT_PREFIX
            "CCTI_Limit: " U32D_FMT PRINT_PREFIX
            "%s\n",
            cc_congestion_control_table.CCTI_Limit,
            out_ss.str().c_str());
}

#if 0
void CongestionControlManager::DumpCCTIList(const CCTIEntryListVec &ccti_list)
{
    struct CC_CongestionControlTable cc_congestion_control_table;
    for (unsigned int idx = 0; idx < ccti_list.size(); idx++) {
        CLEAR_STRUCT(cc_congestion_control_table);
        cc_congestion_control_table.CCTI_Limit =
            (idx * ccti_entry_list_elements_num) + (ccti_entry_list_elements_num - 1);
        cc_congestion_control_table.CCTI_Entry_List = ccti_list[idx];
        DumpCCTableSetting(cc_congestion_control_table);
    }
}
#endif

int CongestionControlManager::GenerateCCTI(const unsigned int control_table_cap,
        CCTIEntryListVec& ccti_list)
{
    int rc = 0;
    ccti_list.clear();
    //Allocate table for current control_table value
    for (unsigned int idx = 0; idx < control_table_cap; idx++) {
        struct CCTI_Entry_List ccti_entry_list;
        CLEAR_STRUCT(ccti_entry_list);
        ccti_list.push_back(ccti_entry_list);
    }
    //we calculate the table assuming it is quadratic and V[95] = nHosts
    //K = nHosts / (95^2)
    double K = 1.0 * m_num_hosts / 9025;

    int divider;
    for (unsigned int idx = 0; idx < ccti_entry_list_elements_num*control_table_cap; idx++) {
        double value = K * idx * idx;
        if (value > CC_GENERAL_DEFAULT_MAX_MANTICE) {
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                    "CCT value is out of CCT max range ( %f )\n", value);
            value = CC_GENERAL_DEFAULT_MAX_MANTICE;
            divider = 0;
        }
        else {
            // check to see if there is a need to use smaller divider
            for (divider = CC_GENERAL_DEFAULT_DIVIDER;
                 divider &&
                 value > (CC_GENERAL_DEFAULT_MAX_MANTICE >> divider);
                 divider--) {
            }
        }
        int multiplier = (int) (value * pow(2,divider));
        ccti_list[idx / ccti_entry_list_elements_num].
            CCTI_Entry_ListElement[idx % ccti_entry_list_elements_num].CCT_Multiplier = multiplier;
        ccti_list[idx / ccti_entry_list_elements_num].
            CCTI_Entry_ListElement[idx % ccti_entry_list_elements_num].CCT_Shift = divider;

    }
    return rc;
}

int CongestionControlManager::GetDefaultCANodeCCTIList(const unsigned int control_table_cap,
        CCTIEntryListVec& ccti_list)
{
    int rc = 0;
    if (m_ca_cc_setting_db.m_default_ccti_list.find(CC_GENERAL_DEFAULT_CCTI_LIST) !=
        m_ca_cc_setting_db.m_default_ccti_list.end() &&
        m_ca_cc_setting_db.m_default_ccti_list[CC_GENERAL_DEFAULT_CCTI_LIST].size() &&
        m_ca_cc_setting_db.m_default_ccti_list[CC_GENERAL_DEFAULT_CCTI_LIST].size() <= control_table_cap) {
        ccti_list = m_ca_cc_setting_db.m_default_ccti_list[CC_GENERAL_DEFAULT_CCTI_LIST];
        return rc;
    }
    if (m_ca_cc_setting_db.m_default_ccti_list.find(control_table_cap) ==
        m_ca_cc_setting_db.m_default_ccti_list.end()) {
        m_ca_cc_setting_db.m_default_ccti_list[control_table_cap] = CCTIEntryListVec();
        rc = GenerateCCTI(control_table_cap, m_ca_cc_setting_db.m_default_ccti_list[control_table_cap]);
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Calculated CCT for control_table_cap: %d\n", control_table_cap);
    }
    ccti_list = m_ca_cc_setting_db.m_default_ccti_list[control_table_cap];
    return rc;
}

int CongestionControlManager::GetCANodeCCTIList(const CCNodeInfo& node_info,
        bool& node_supports_cc,
        CCTIEntryListVec& ccti_list,
        map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator& ca_node_it)
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_ca_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
        m_ca_cc_setting_db.m_nodes_db.end()) {
        // No information found for this node
        return 1;
    }
    ca_node_it = node_it;
    node_supports_cc = (node_it->second.m_cc_support == CC_SUPPORTED);
    if (!node_supports_cc) {
        return rc;
    }
    // When ccti list isn't empty, it was configured by user
    if (!node_it->second.m_node_ccti_list.size()) {
        // CCTI list is empty - need to create default list for current control_table_cap
        // If m_control_table_cap equals zero, use default value.
        // Node value wasn't retrieved successfully.
        rc = GetDefaultCANodeCCTIList(node_it->second.m_control_table_cap ?
                                      node_it->second.m_control_table_cap :
                                      CA_CONG_SETTING_DEFAULT_CONTROL_TABLE_CAP,
                                      node_it->second.m_node_ccti_list);
    }
    ccti_list = node_it->second.m_node_ccti_list;
    return rc;
}

bool CongestionControlManager::CompareCACongestionControlTableSetting(const CCTIEntryListVec &lhs,
        const CCTIEntryListVec &rhs)
{
    if (lhs.size() != rhs.size()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "CCTIEntryListVec size mismatch. %zu : %zu\n",
            lhs.size(), rhs.size());
        return false;
    }
    for (unsigned int entry = 0; entry < lhs.size(); entry++) {
        for (unsigned int idx = 0; idx < ccti_entry_list_elements_num; idx++) {
            if (lhs[entry].CCTI_Entry_ListElement[idx].CCT_Multiplier !=
                rhs[entry].CCTI_Entry_ListElement[idx].CCT_Multiplier) {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                        "CCTI_Entry_ListElement[%d].CCT_Multiplier mismatch. %d : %d\n",
                        (entry*ccti_entry_list_elements_num)+ idx,
                        lhs[entry].CCTI_Entry_ListElement[idx].CCT_Multiplier,
                        rhs[entry].CCTI_Entry_ListElement[idx].CCT_Multiplier);
                return false;
            }

            if (lhs[entry].CCTI_Entry_ListElement[idx].CCT_Shift !=
                rhs[entry].CCTI_Entry_ListElement[idx].CCT_Shift) {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                        "CCTI_Entry_ListElement[%d].CCT_Shift mismatch. %d : %d\n",
                        (entry*ccti_entry_list_elements_num)+ idx,
                        lhs[entry].CCTI_Entry_ListElement[idx].CCT_Shift,
                        rhs[entry].CCTI_Entry_ListElement[idx].CCT_Shift);
                return false;
            }
        }
    }
    return true;
}

int CongestionControlManager::SetCACongestionControlTable(const CCNodeInfo& node_info)
{
    int rc = 0;
    struct CC_CongestionControlTable cc_congestion_control_table;
    CCTIEntryListVec ccti_list;
    bool node_supports_cc = false;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    rc = GetCANodeCCTIList(node_info, node_supports_cc, ccti_list, ca_node_it);
    if (rc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Node with GUID " U64H_FMT " can't be configured. "
                "Congestion Control information can't be retrieved from"
                " the node\n", node_info.m_port_guid);
        return rc;
    }
    if (!node_supports_cc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Node with GUID " U64H_FMT
                 " doesn't support Congestion Control\n",
                node_info.m_port_guid);
        return rc;
    }
    for (unsigned int idx = 0; idx < ccti_list.size(); idx++) {
        CLEAR_STRUCT(cc_congestion_control_table);
        cc_congestion_control_table.CCTI_Limit =
            (idx * ccti_entry_list_elements_num) + (ccti_entry_list_elements_num - 1);
        cc_congestion_control_table.CCTI_Entry_List = ccti_list[idx];
        rc = m_ibis_obj.CCCongestionControlTableSet(node_info.m_lid,
                                                    node_info.m_sl,
                                                    m_cc_key, idx,
                                                    &cc_congestion_control_table);
        CheckRC(rc);
        if (!rc) {
            ca_node_it->second.m_is_control_table_set = true;
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                    "Set CA node with GUID " U64H_FMT
                    " with the following CCTable parameters:\n", node_info.m_port_guid);
            DumpCCTableSetting(cc_congestion_control_table);
        }
        else {
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Failed to send CongestionControlTable [Set] to node GUID:"
                    U64H_FMT "\n", node_info.m_port_guid);
            if (m_enable == false) {
                ca_node_it->second.m_error_count++;
            }
        }
    }
    return rc;
}

void CongestionControlManager::SetCACongestionControlTableForAllNodes()
{
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()){
        if (ca_node_it->second.m_cc_support == CC_SUPPORTED) {
            if (!ca_node_it->second.m_is_control_table_set ||
                ca_node_it->second.m_osm_update_needed) {

                if (m_enable || ca_node_it->second.m_error_count < MAX_DISABLE_ERROR_COUNT) {                
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                            "CA node with GUID " U64H_FMT ". "
                            PRINT_PREFIX "m_osm_update_needed: %d, m_is_control_table_set: %d\n",
                            ca_node_it->second.m_node_info.m_port_guid,
                            ca_node_it->second.m_osm_update_needed,
                            ca_node_it->second.m_is_control_table_set);
    
                    SetCACongestionControlTable(ca_node_it->second.m_node_info);
                } else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                            "Do not set CongestionControlTable for CA node with GUID " U64H_FMT
                            "because of previous disable error.\n", ca_node_it->second.m_node_info.m_port_guid);
                }
            } else {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                        "No need to set CongestionControlTable for CA node with GUID " U64H_FMT
                        "\n", ca_node_it->second.m_node_info.m_port_guid);
            }
        }
        ca_node_it++;
    }
    return;
}


/***************/
void CongestionControlManager::DumpSWCongSetting(struct CC_SwitchCongestionSetting& cc_sw_congestion_setting)
{
    string victim_str = DumpMask("Victim_Mask", cc_sw_congestion_setting.Victim_Mask);
    string credit_str = DumpMask("Credit_Mask", cc_sw_congestion_setting.Credit_Mask);
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "\n" PRINT_PREFIX
            "Packet_Size: 0x%02x, Threshold: 0x%02x, Marking_Rate: 0x%04x, "
            "CS_Threshold: 0x%02x, CS_ReturnDelay: 0x%04x" PRINT_PREFIX
            "%s"
            "%s\n",
            cc_sw_congestion_setting.Packet_Size,
            cc_sw_congestion_setting.Threshold,
            cc_sw_congestion_setting.Marking_Rate,
            cc_sw_congestion_setting.CS_Threshold,
            cc_sw_congestion_setting.CS_ReturnDelay,
            victim_str.c_str(),
            credit_str.c_str());
}

void CongestionControlManager::SetMask(UINT256& mask, u_int8_t bit)
{
    switch (bit) {
    case 0 ... 31 :
        mask.Mask_31_0 |= 1 << (bit - 0);
        break;
    case 32 ... 63 :
        mask.Mask_63_32 |= 1 << (bit - 32);
        break;
    case 64 ... 95 :
        mask.Mask_95_64 |= 1 << (bit - 64);
        break;
    case 96 ... 127 :
        mask.Mask_127_96 |= 1 << (bit - 96);
        break;
    case 128 ... 159 :
        mask.Mask_159_128 |= 1 << (bit - 128);
        break;
    case 160 ... 191 :
        mask.Mask_191_160 |= 1 << (bit - 160);
        break;
    case 192 ... 223 :
        mask.Mask_223_192 |= 1 << (bit - 192);
        break;
    case 224 ... 255 :
        mask.Mask_255_224 |= 1 << (bit - 224);
        break;
    }
}

int CongestionControlManager::GenerateVictimMaskList(const CCNodeInfo& node_info,
        UINT256& victimMaskList)
{
    int rc = 0;
    uint8_t remote_port_num;
    CLEAR_STRUCT(victimMaskList);
    for (unsigned int port_num = 1; port_num <
          osm_node_get_num_physp(node_info.m_p_osm_node); port_num++) {
        const osm_physp_t *p_osm_port =
            osm_node_get_physp_ptr(node_info.m_p_osm_node, port_num);
        if (!p_osm_port || !osm_link_is_healthy(p_osm_port)) {
            continue;
        }

        const osm_physp_t *p_remote_osm_port = osm_physp_get_remote(p_osm_port);
        if (!p_remote_osm_port) {
            continue;
        }

        const osm_node_t *p_remote_node = osm_node_get_remote_node(node_info.m_p_osm_node,
                                                                   port_num, &remote_port_num);

        /* ignore any loopback connection on switch */
        if (node_info.m_p_osm_node == p_remote_node) {
            continue;
        }

        uint8_t remote_node_type = osm_node_get_type(p_remote_node);

        switch (remote_node_type) {
        case IB_NODE_TYPE_CA:
            /* switch connected to hca */
            SetMask(victimMaskList,port_num);
            break;

        case IB_NODE_TYPE_ROUTER:
            /* leaving this port - proceeding to the next one */
            continue;

        case IB_NODE_TYPE_SWITCH:
            /* switch connected to another switch */
            continue;

        default:
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Node GUID " U64D_FMT
                    " - Unknown node type: %s\n",
                    cl_ntoh64(osm_node_get_node_guid(p_remote_node)),
                    ib_get_node_type_str(remote_node_type));
            return 1;
        }
    }
    return rc;
}

int CongestionControlManager::GetSWNodeCCSettings(const CCNodeInfo& node_info,
        bool& node_supports_cc,
        struct CC_SwitchCongestionSetting& node_params,
        struct CC_SwitchPortCongestionSetting port_params[NUM_OF_32_PORTS],
        map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator& sw_node_it)
{
    int rc = 0;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator node_it;
    if ((node_it = m_sw_cc_setting_db.m_nodes_db.find(node_info.m_port_guid)) ==
        m_sw_cc_setting_db.m_nodes_db.end()) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to find Switch node with GUID:" U64H_FMT "\n",
                node_info.m_port_guid);
        return 1;
    }
    sw_node_it = node_it;
    node_supports_cc = (node_it->second.m_cc_support == CC_SUPPORTED);
    if (!node_supports_cc) {
        return rc;
    }
    node_params = node_it->second.m_node_params;
    for (unsigned int idx = 0; idx < NUM_OF_32_PORTS; idx++) {
        port_params[idx] = node_it->second.m_port_params[idx];
    }
    return rc;
}

bool CongestionControlManager::CompareSwitchCongestionSetting(const struct CC_SwitchCongestionSetting &lhs,
        const struct CC_SwitchCongestionSetting &rhs)
{
    if (lhs.Control_Map.Credit_MaskIsValid !=
        rhs.Control_Map.Credit_MaskIsValid) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Control_Map.Credit_MaskIsValid mismatch. %d : %d\n",
            lhs.Control_Map.Credit_MaskIsValid,
            rhs.Control_Map.Credit_MaskIsValid);
        return false;
    }
    if (lhs.Control_Map.CS_ThresholdAndCS_ReturnDelayIsValid !=
        rhs.Control_Map.CS_ThresholdAndCS_ReturnDelayIsValid) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Control_Map.CS_ThresholdAndCS_ReturnDelayIsValid mismatch. %d : %d\n",
            lhs.Control_Map.CS_ThresholdAndCS_ReturnDelayIsValid,
            rhs.Control_Map.CS_ThresholdAndCS_ReturnDelayIsValid);
        return false;
    }
    if (lhs.Control_Map.Marking_RateIsValid !=
        rhs.Control_Map.Marking_RateIsValid) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Control_Map.Marking_RateIsValid mismatch. %d : %d\n",
            lhs.Control_Map.Marking_RateIsValid,
            rhs.Control_Map.Marking_RateIsValid);
        return false;
    }
    if (lhs.Control_Map.ThresholdAndPacket_SizeIsValid !=
        rhs.Control_Map.ThresholdAndPacket_SizeIsValid) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch ThresholdAndPacket_SizeIsValid mismatch. %d : %d\n",
            lhs.Control_Map.ThresholdAndPacket_SizeIsValid,
            rhs.Control_Map.ThresholdAndPacket_SizeIsValid);
        return false;
    }
    if (lhs.Control_Map.Victim_MaskIsValid !=
        rhs.Control_Map.Victim_MaskIsValid) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Control_Map.Victim_MaskIsValid mismatch. %d : %d\n",
            lhs.Control_Map.Victim_MaskIsValid,
            rhs.Control_Map.Victim_MaskIsValid);
        return false;
    }
    if (lhs.Credit_Mask.Mask_31_0 != rhs.Credit_Mask.Mask_31_0) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Credit_Mask.Mask_31_0 mismatch. %d : %d\n",
            lhs.Credit_Mask.Mask_31_0,
            rhs.Credit_Mask.Mask_31_0);
        return false;
    }
    if (lhs.Credit_Mask.Mask_63_32 != rhs.Credit_Mask.Mask_63_32) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Credit_Mask.Mask_63_32 mismatch. %d : %d\n",
            lhs.Credit_Mask.Mask_63_32,
            rhs.Credit_Mask.Mask_63_32);
        return false;
    }
    if (lhs.CS_ReturnDelay != rhs.CS_ReturnDelay) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch CS_ReturnDelay mismatch. %d : %d\n",
            lhs.CS_ReturnDelay,
            rhs.CS_ReturnDelay);
        return false;
    }
    if (lhs.CS_Threshold != rhs.CS_Threshold) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch CS_Threshold mismatch. %d : %d\n",
            lhs.CS_Threshold,
            rhs.CS_Threshold);
        return false;
    }
    if (lhs.Marking_Rate != rhs.Marking_Rate) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Marking_Rate mismatch. %d : %d\n",
            lhs.Marking_Rate,
            rhs.Marking_Rate);
        return false;
    }
    if (lhs.Packet_Size != rhs.Packet_Size) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Packet_Size mismatch. %d : %d\n",
            lhs.Packet_Size,
            rhs.Packet_Size);
        return false;
    }
    if (lhs.Threshold != rhs.Threshold) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Threshold mismatch. %d : %d\n",
            lhs.Threshold,
            rhs.Threshold);
        return false;
    }
    if (lhs.Victim_Mask.Mask_31_0 != rhs.Victim_Mask.Mask_31_0) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Victim_Mask.Mask_31_0 mismatch. %d : %d\n",
            lhs.Victim_Mask.Mask_31_0,
            rhs.Victim_Mask.Mask_31_0);
        return false;
    }
    if (lhs.Victim_Mask.Mask_63_32 != rhs.Victim_Mask.Mask_63_32) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
            "Switch Victim_Mask.Mask_63_32 mismatch. %d : %d\n",
            lhs.Victim_Mask.Mask_63_32,
            rhs.Victim_Mask.Mask_63_32);
        return false;
    }
    return true;
}

int CongestionControlManager::SetSwitchCongestionSetting(const CCNodeInfo& node_info)
{
    int rc = 0;
    struct CC_SwitchCongestionSetting cc_sw_congestion_setting;
    struct CC_SwitchPortCongestionSetting cc_sw_port_congestion_setting[NUM_OF_32_PORTS];
    bool node_supports_cc = false;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    rc = GetSWNodeCCSettings(node_info,
                             node_supports_cc,
                             cc_sw_congestion_setting,
                             cc_sw_port_congestion_setting,
                             sw_node_it);
    if (rc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Node with GUID " U64H_FMT " can't be configured - "
                "CC information can't be retrieved for the node\n",
                node_info.m_port_guid);
        return rc;
    }
    if (!node_supports_cc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Node with GUID " U64H_FMT " doesn't support Congestion Control\n",
                node_info.m_port_guid);
        return rc;
    }
    GenerateVictimMaskList(node_info,cc_sw_congestion_setting.Victim_Mask);
    rc = m_ibis_obj.CCSwitchCongestionSettingSet(node_info.m_lid,
                                                 node_info.m_sl,
                                                 m_cc_key,
                                                 &cc_sw_congestion_setting);
    CheckRC(rc);
    if (!rc) {
        sw_node_it->second.m_is_congestion_settings_set = true;
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Set Switch node with GUID " U64H_FMT
                 " with the following SWCongestionSetting parameters:\n", node_info.m_port_guid);
        DumpSWCongSetting(cc_sw_congestion_setting);
    }
    else {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Failed to send SwitchCongestionSetting [Set] to node GUID:"
                U64H_FMT "\n", node_info.m_port_guid);
        sw_node_it->second.m_error_count++;
    }
    return rc;
}

void CongestionControlManager::SetSwitchCongestionSettingForAllNodes()
{
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    sw_node_it = m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=  m_sw_cc_setting_db.m_nodes_db.end()){
        if (sw_node_it->second.m_cc_support == CC_SUPPORTED) {
            if (!sw_node_it->second.m_is_congestion_settings_set ||
                sw_node_it->second.m_osm_update_needed) {

                if (m_enable || sw_node_it->second.m_error_count < MAX_DISABLE_ERROR_COUNT) {                

                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                            "SW node with GUID " U64H_FMT ". "
                            PRINT_PREFIX "m_osm_update_needed: %d, m_is_congestion_settings_set: %d\n",
                            sw_node_it->second.m_node_info.m_port_guid,
                            sw_node_it->second.m_osm_update_needed,
                            sw_node_it->second.m_is_congestion_settings_set);
    
                    SetSwitchCongestionSetting(sw_node_it->second.m_node_info);
                } else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                            "Do not set disable SwitchCongestionSetting for switch node with GUID " U64H_FMT
                             "because of previous error.\n", sw_node_it->second.m_node_info.m_port_guid);
                }
            } else {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                        "No need to set SwitchCongestionSetting for switch node with GUID " U64H_FMT
                         "\n", sw_node_it->second.m_node_info.m_port_guid);
            }
        }
        sw_node_it++;
    }
    return;
}


/***************/
void CongestionControlManager::DumpSWPortCongSetting(const unsigned int block_idx,
        struct CC_SwitchPortCongestionSetting& cc_sw_port_congestion_setting)
{
    stringstream out_ss;
    char buff[128];
    for (unsigned int port = block_idx*32 ; port < (block_idx+1)*32; port++) {
        CLEAR_STRUCT(buff);
        if (cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Control_Type) {
            sprintf(buff,"[port:%2d] CS_Threshold: 0x%02x, Valid: " U32D_FMT "\n",
                    port,
                    cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Threshold,
                    cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Valid);
        }
        else {
            sprintf(buff,"[port:%2d] Marking Rate: 0x%04x, Packet_Size: 0x%02x, "
                    "Threshold: 0x%02x, Valid: " U32D_FMT "\n",
                    port,
                    cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Cong_Parm,
                    cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Packet_Size,
                    cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Threshold,
                    cc_sw_port_congestion_setting.SwitchPortCongestionSettingElement[port%32].Valid);

        }
        out_ss << buff;
    }
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,PRINT_PREFIX
            "%s\n", out_ss.str().c_str());
}

int CongestionControlManager::SetSwitchPortCongestionSetting(const CCNodeInfo& node_info)
{
    int rc = 0;
    struct CC_SwitchCongestionSetting cc_sw_congestion_setting;
    struct CC_SwitchPortCongestionSetting cc_sw_port_congestion_setting[NUM_OF_32_PORTS];
    bool node_supports_cc = false;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    rc = GetSWNodeCCSettings(node_info,
                             node_supports_cc,
                             cc_sw_congestion_setting,
                             cc_sw_port_congestion_setting,
                             sw_node_it);
    if (rc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Node with GUID " U64H_FMT " can't be configured - "
                "CC information can't be retrieved\n",
                node_info.m_port_guid);
        return rc;
    }
    if (!node_supports_cc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                "Node with GUID " U64H_FMT
                 " doesn't support Congestion Control\n", node_info.m_port_guid);
        return rc;
    }

    for (unsigned int block_idx = 0; block_idx < NUM_OF_32_PORTS; block_idx++) {
        int valid = 0;
        for (unsigned int port = 0; port < 32; port++) {
            valid |= cc_sw_port_congestion_setting[block_idx].
                SwitchPortCongestionSettingElement[port].Valid;
        }
        if (!valid) {
            sw_node_it->second.m_is_port_congestion_settings_set &= true;
            continue;
        }
        rc = m_ibis_obj.CCSwitchPortCongestionSettingSet(node_info.m_lid,
                                                         node_info.m_sl,
                                                         m_cc_key,
                                                         block_idx,
                                                         &cc_sw_port_congestion_setting[block_idx]);
        CheckRC(rc);
        if (!rc) {
            sw_node_it->second.m_is_port_congestion_settings_set &= true;
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                    "Set Switch node with GUID " U64H_FMT
                     " with the following SWPortCongSetting parameters:\n", node_info.m_port_guid);
            DumpSWPortCongSetting(block_idx, cc_sw_port_congestion_setting[block_idx]);
        }
        else {
            sw_node_it->second.m_is_port_congestion_settings_set &= false;
            OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                    "Failed to send SwitchPortCongestionSetting "
                    "[Set] to node GUID:" U64H_FMT "\n", node_info.m_port_guid);

            sw_node_it->second.m_error_count++;

        }
    }

    return rc;
}

void CongestionControlManager::SetSwitchPortCongestionSettingForAllNodes()
{
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    sw_node_it = m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=  m_sw_cc_setting_db.m_nodes_db.end()){
        if (sw_node_it->second.m_cc_support == CC_SUPPORTED) {
            if (!sw_node_it->second.m_is_port_congestion_settings_set ||
                sw_node_it->second.m_osm_update_needed) {

                if (m_enable || sw_node_it->second.m_error_count < MAX_DISABLE_ERROR_COUNT) {                

                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
                            "SW node with GUID " U64H_FMT ". "
                            PRINT_PREFIX "m_osm_update_needed: %d, m_is_port_congestion_settings_set: %d\n",
                            sw_node_it->second.m_node_info.m_port_guid,
                            sw_node_it->second.m_osm_update_needed,
                            sw_node_it->second.m_is_port_congestion_settings_set);
    
                    SetSwitchPortCongestionSetting(sw_node_it->second.m_node_info);
                } else {
                    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                            "Do not set disable SwitchPortCongestionSetting for switch node with GUID " U64H_FMT
                             "because of previous error.\n", sw_node_it->second.m_node_info.m_port_guid);
                }
            } else {
                OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
                        "No need to set SwitchPortCongestionSetting for switch node with GUID " U64H_FMT
                         "\n", sw_node_it->second.m_node_info.m_port_guid);
            }
        }
        sw_node_it++;
    }
}


/***************/
void CongestionControlManager::SetCCParamsForAllNodes()
{
    GetCAControlTableCapForAllNodes();

    SetCACongestionSettingForAllNodes();
    SetCACongestionControlTableForAllNodes();

    SetSwitchCongestionSettingForAllNodes();
    SetSwitchPortCongestionSettingForAllNodes();
    return;
}


//============================================================================//
/***************/
string CongestionControlManager::DumpMask(const string &mask_name, const UINT256 &mask_st)
{
    stringstream ports_ss;
    stringstream output_ss;

    u_int32_t mask;
    for (unsigned int port = 0; port < 64; port++) {
        mask = (port < 32) ?    mask_st.Mask_31_0 : mask_st.Mask_63_32;
        if (mask & (1 << port)) {
            ports_ss << std::dec << port << ", ";
        }
    }
    output_ss << mask_name << ": " << ports_ss.str() << PRINT_PREFIX;
    return output_ss.str();
}

string CongestionControlManager::DumpCACongestionLog(struct CC_CongestionLogCA& cc_congestion_log_ca)
{
    char buff[512];
    stringstream output_ss;
    sprintf(buff,"LogType: %d, CongestionFlags: %s, ThresholdCongestionEventMap: " U32H_FMT
                 ", CurrentTimeStamp: " U32H_FMT "\n\n",
                 cc_congestion_log_ca.LogType,
                 (cc_congestion_log_ca.CongestionFlags & 0x1) ?
                 "CC_Key lease period timer active" :
                 "CC_Key lease period timer inactive",
                 cc_congestion_log_ca.ThresholdCongestionEventMap,
                 cc_congestion_log_ca.CurrentTimeStamp);

    output_ss << buff;

    /*
    qsort(cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement,
          CA_MAX_CC_LOG_EVENT,
          sizeof(CongestionLogEventListCAElement),
          cmp_ca_log_event);
    */

    for (unsigned int log_event = 0; log_event < CA_MAX_CC_LOG_EVENT; log_event++) {
        if (!cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].Timestamp) {
            continue;
        }
        string service_type_cn_entry;
        switch (cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].Service_Type_CN_Entry) {
        case 0:
            service_type_cn_entry = "RC";
            break;
        case 1:
            service_type_cn_entry = "UC";
            break;
        case 2:
            service_type_cn_entry = "RD";
            break;
        case 3:
            service_type_cn_entry = "UD";
            break;
        default:
            service_type_cn_entry = "UNKNOWN";
        }
        sprintf(buff,"\t[event:%2d] LQP: " U32H_FMT
                               ", RQP: " U32H_FMT
                               ", SL: %2u"
                               ", ST: %s"
                               ", DLID: %5d"
                               ", Timestamp: " U32H_FMT,
                log_event,
                cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].Local_QP_CN_Entry,
                cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].Remote_QP_Number_CN_Entry,
                cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].SL_CN_Entry,
                service_type_cn_entry.c_str(),
                cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].Remote_LID_CN_Entry,
                cc_congestion_log_ca.CongestionLogEvent.
                    CongestionLogEventListCAElement[log_event].Timestamp);

        output_ss << buff << "\n";
    }
    return output_ss.str();
}

string CongestionControlManager::DumpSWCongestionLog(struct CC_CongestionLogSwitch& cc_congestion_log_sw)
{
    char buff[512];
    stringstream output_ss;

    string port_map_str = DumpMask("PortMap", cc_congestion_log_sw.PortMap);
    sprintf(buff,"LogType: %d, CongestionFlags: %s, CurrentTimeStamp: " U32H_FMT ", %s\n",
                 cc_congestion_log_sw.LogType,
                 (cc_congestion_log_sw.CongestionFlags & 0x1) ?
                 "CC_Key lease period timer active" :
                 "CC_Key lease period timer inactive",
                 cc_congestion_log_sw.CurrentTimeStamp,
                 port_map_str.c_str());

    output_ss << buff;

    /*
    qsort(cc_congestion_log_sw.CongestionEntryList.
                    CongestionEntryListSwitchElement,
          SW_MAX_CC_LOG_EVENT,
          sizeof(CongestionEntryListSwitchElement),
          cmp_sw_log_event);
    */

    for (unsigned int log_event = 0; log_event < SW_MAX_CC_LOG_EVENT; log_event++) {
        if (!cc_congestion_log_sw.CongestionEntryList.
                    CongestionEntryListSwitchElement[log_event].Timestamp) {
            continue;
        }
        sprintf(buff,"\t[event:%2d] SLID: %5u"
                               ", DLID: %5u"
                               ", SL: %2u"
                               ", Timestamp: " U32H_FMT,
                log_event,
                cc_congestion_log_sw.CongestionEntryList.
                    CongestionEntryListSwitchElement[log_event].SLID,
                cc_congestion_log_sw.CongestionEntryList.
                    CongestionEntryListSwitchElement[log_event].DLID,
                cc_congestion_log_sw.CongestionEntryList.
                    CongestionEntryListSwitchElement[log_event].SL,
                cc_congestion_log_sw.CongestionEntryList.
                    CongestionEntryListSwitchElement[log_event].Timestamp);

        output_ss << buff << "\n";
    }
    return output_ss.str();
}

/***************/
int CongestionControlManager::GetSWCCStatistics(CCNodeInfo& node_info)
{
    map<u_int64_t, u_int16_t>::iterator sw_node_it;
    struct CC_CongestionLogSwitch cc_congestion_log_sw;
    int rc = m_ibis_obj.CCCongestionLogSwitchGet(node_info.m_lid, node_info.m_sl, m_cc_key,
                                                 &cc_congestion_log_sw);
    if (rc) {
        CC_MGR_LOG(TT_LOG_LEVEL_ERROR,"\n\n"
                   " Failed to send CongestionLogSwitch [Get] to node GUID:" U64H_FMT "\n",
                   node_info.m_port_guid);
        return rc;
    }

    sw_node_it = m_sw_cc_log_event_cnt_db.begin();
    if ((sw_node_it = m_sw_cc_log_event_cnt_db.find(node_info.m_port_guid)) ==
                      m_sw_cc_log_event_cnt_db.end()) {
        m_sw_cc_log_event_cnt_db[node_info.m_port_guid] = 0;
        sw_node_it = m_sw_cc_log_event_cnt_db.find(node_info.m_port_guid);
    }
    if (cc_congestion_log_sw.LogEventsCounter > sw_node_it->second) {
        m_cc_statistics_db.m_num_sw_ports_congested += GetSWNumCongPorts(node_info,
                                                                         cc_congestion_log_sw);
        string event_dump = DumpSWCongestionLog(cc_congestion_log_sw);
        CC_MGR_LOG(TT_LOG_LEVEL_INFO,"\n\n"
                   " Switch with GUID: " U64H_FMT ", has %d new congestion control events\n %s\n\n",
                   node_info.m_port_guid,
                   cc_congestion_log_sw.LogEventsCounter - sw_node_it->second,
                   event_dump.c_str());
    }
    sw_node_it->second = cc_congestion_log_sw.LogEventsCounter;
    return rc;
}

int CongestionControlManager::GetCACCStatistics(CCNodeInfo& node_info)
{
    map<u_int64_t, u_int16_t>::iterator ca_node_it;
    struct CC_CongestionLogCA cc_congestion_log_ca;
    int rc = m_ibis_obj.CCCongestionLogCAGet(node_info.m_lid, node_info.m_sl, m_cc_key,
                                             &cc_congestion_log_ca);
    if (rc) {
        CC_MGR_LOG(TT_LOG_LEVEL_ERROR,"\n\n"
                   " Failed to send CongestionLogCA [Get] to node GUID:" U64H_FMT "\n",
                   node_info.m_port_guid);
        return rc;
    }

    ca_node_it = m_ca_cc_log_event_cnt_db.begin();
    if ((ca_node_it = m_ca_cc_log_event_cnt_db.find(node_info.m_port_guid)) ==
                      m_ca_cc_log_event_cnt_db.end()) {
        m_ca_cc_log_event_cnt_db[node_info.m_port_guid] = 0;
        ca_node_it = m_ca_cc_log_event_cnt_db.find(node_info.m_port_guid);
    }
    if (cc_congestion_log_ca.ThresholdEventCounter > ca_node_it->second) {
        string event_dump = DumpCACongestionLog(cc_congestion_log_ca);
        CC_MGR_LOG(TT_LOG_LEVEL_INFO,"\n\n"
                   " CA with GUID: " U64H_FMT ", has %d new congestion control events\n %s\n\n",
                   node_info.m_port_guid,
                   cc_congestion_log_ca.ThresholdEventCounter - ca_node_it->second,
                   event_dump.c_str());
    }
    ca_node_it->second = cc_congestion_log_ca.ThresholdEventCounter;
    return rc;
}

unsigned int CongestionControlManager::GetSWNumCongPorts(CCNodeInfo& node_info,
        struct CC_CongestionLogSwitch &cc_congestion_log_sw)
{
    unsigned int num_sw_cong_ports = 0;
    u_int32_t mask;
    for (unsigned int port = 0; port < node_info.m_p_osm_node->node_info.num_ports; port++) {
        mask = (port < 32) ?    cc_congestion_log_sw.PortMap.Mask_31_0 :
                                cc_congestion_log_sw.PortMap.Mask_63_32;
        if (mask & (1 << port)) {
            num_sw_cong_ports++;
        }
    }
    return num_sw_cong_ports;
}

/***************/
int CongestionControlManager::CollectCCStatisticsFromAllCAs()
{
    int rc = 0;
    map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator ca_node_it;
    ca_node_it = m_ca_cc_setting_db.m_nodes_db.begin();
    while (ca_node_it !=  m_ca_cc_setting_db.m_nodes_db.end()){
        if (ca_node_it->second.m_cc_support == CC_SUPPORTED &&
            ca_node_it->second.m_enable) {
            if (CC_STATISTICS_AGENT_SHOULD_EXIT) {
                throw 1;
            }
            GetCACCStatistics(ca_node_it->second.m_node_info);
        }
        ca_node_it++;
    }
    return rc;
}

int CongestionControlManager::CollectCCStatisticsFromAllSwitches()
{
    int rc = 0;
    map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator sw_node_it;
    m_cc_statistics_db.m_num_sw_ports_congested = 0;
    sw_node_it = m_sw_cc_setting_db.m_nodes_db.begin();
    while (sw_node_it !=  m_sw_cc_setting_db.m_nodes_db.end()){
        if (sw_node_it->second.m_cc_support == CC_SUPPORTED &&
            sw_node_it->second.m_enable) {
            if (CC_STATISTICS_AGENT_SHOULD_EXIT) {
                throw 1;
            }
            GetSWCCStatistics(sw_node_it->second.m_node_info);
        }
        sw_node_it++;
    }
    return rc;
}

/***************/
void CongestionControlManager::CCStatisticsSleep(time_t seconds_to_sleep)
{
    time_t end_time = time(NULL) + seconds_to_sleep;

    // We could just sleep for seconds_to_sleep, but we
    // want to be able to interrupt this sleep when required.
    while (time(NULL) < end_time) {
        if (CC_STATISTICS_AGENT_SHOULD_EXIT)
            return;
        sleep(1);
    }
    return;
}

int CongestionControlManager::ReportCCStatistics()
{
    int rc = 0;
    rc = ReportFit(FIT_REPORT_CONGESTED_SWITCH_PORTS, 1, &m_cc_statistics_db.m_num_sw_ports_congested);
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE,
            "Number of congested switch ports: " U64D_FMT "\n",
            m_cc_statistics_db.m_num_sw_ports_congested);
    return rc;
}

void CongestionControlManager::RunCCStatisticsAgent()
{
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_INFO, "Congestion Control Statistics Agent started\n");

    while (m_run_cc_statistics_agent) {
        pthread_mutex_lock(&m_cc_gate_mutex);
        StatLockCCMutex();
        pthread_mutex_unlock(&m_cc_gate_mutex);

        CC_MGR_LOG(TT_LOG_LEVEL_INFO ,"\n\n"
                " Started Congestion Control Statistics cycle\n\n");
        try {
            CollectCCStatisticsFromAllCAs();
            CollectCCStatisticsFromAllSwitches();

            ReportCCStatistics();

            CC_MGR_LOG(TT_LOG_LEVEL_INFO, "\n\n"
                    " Finished Congestion Control Statistics cycle\n\n");
        } catch (int) {
            CC_MGR_LOG(TT_LOG_LEVEL_INFO,"\n\n"
                    " Congestion Control Statistics aborted due to CC MGR request\n\n");
        }

        UnLockCCMutex(LOCK_HELD_BY_STAT);

        CCStatisticsSleep((time_t)m_cc_statistics_cycle);
    }
}

/***************/
void CongestionControlManager::CreateCCStatisticsAgent()
{
    int rc = 0;
    if (!m_cc_statistics_cycle || !m_cc_config_up) {
        // Create CC Statistics Agent only if fabric successfully configured and
        // cycle greater than zero.
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG, "CC Statistics Agent will not be created."
                PRINT_PREFIX "m_cc_statistics_cycle: %ld, m_cc_config_up: %d\n",
                m_cc_statistics_cycle, m_cc_config_up);
        return;
    }
    if (m_cc_log_agent_created == true) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG, "CC Statistics Agent already exists\n");
        return;
    }

    m_run_cc_statistics_agent = true;
    rc = pthread_create(&m_cc_log_agent_thread, 0, &CongestionControlManager::run_cc_statistics_agent, this);
    if (rc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR, "CC Statistics Agent wasn't created. Error: %d\n", rc);
        return;
    }

    m_cc_log_agent_created = true;
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "CC Statistics Agent created\n");
    return;
}

void CongestionControlManager::DestroyCCStatisticsAgent()
{
    if (m_cc_log_agent_created == false) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_VERBOSE, "CC Statistics Agent wasn't created. No need to destroy\n");
        return;
    }
    UnLockCCMutex(LOCK_HELD_BY_MGR);
    m_cc_mgr_request = true;
    m_run_cc_statistics_agent = false;
    pthread_join(m_cc_log_agent_thread,0);
}


//============================================================================//
int CongestionControlManager::ReportFit(fit_msg_type_t msg_type, const int cnt, void* msg_data)
{
    do_smd_report(msg_type, cnt, msg_data);
    return 0;
}


/****************************************************/
/**************** public methods ********************/
/****************************************************/
//============================================================================//
CongestionControlManager::CongestionControlManager(osm_opensm_t *p_osm,
        char *osm_plugin_options) :
        m_num_hosts(CC_GENERAL_DEFAULT_MIN_HOSTS), m_port_guid(0),
        m_p_osm(p_osm),
        m_p_osm_vendor(p_osm->p_vendor),
        m_p_osm_subn(&p_osm->subn),
        m_p_osm_sa(&p_osm->sa),
        m_p_osm_log(&p_osm->log),
        m_conf_file_name(CONF_INPUT_FILE_NAME),
        m_max_errors(CC_GENERAL_DEFAULT_MAX_ERRORS),
        m_error_window(CC_GENERAL_DEFAULT_ERROR_WINDOW), m_p_error_window(NULL),
        m_cc_config_up(false), m_cc_log_agent_created(false),
        m_run_cc_statistics_agent(false), m_cc_mgr_request(true),
        m_cc_statistics_cycle(CC_GENERAL_DEFAULT_CC_STATISTICS_CYCLE)
{
    Construct(osm_plugin_options);
}


//============================================================================//
CongestionControlManager::~CongestionControlManager()
{
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_INFO, "Exiting...\n");
    if (m_p_error_window) {
        delete[] m_p_error_window;
    }
    DestroyCCStatisticsAgent();
    tt_log_destroy();
    pthread_mutex_destroy(&m_cc_mutex);
    pthread_mutex_destroy(&m_cc_gate_mutex);
    pthread_mutex_destroy(&m_lock_holder_mutex);
}


//============================================================================//
void CongestionControlManager::SetOSMEventId(osm_epi_event_id_t osm_event_id)
{
    m_osm_event_id = osm_event_id;
    switch (m_osm_event_id) {
    case OSM_EVENT_ID_SUBNET_UP:
    case OSM_EVENT_ID_HEAVY_SWEEP_DONE:
    case OSM_EVENT_ID_STATE_CHANGE:
        MgrLockCCMutex();
        break;
    default:
        break;
    }
}


//============================================================================//
int CongestionControlManager::UpdateFabricNodes()
{
    int rc = 0;
    osm_port_t *p_osm_port;
    osm_port_t *p_next_osm_port;

    p_next_osm_port = (osm_port_t *)cl_qmap_head(&m_p_osm_subn->port_guid_tbl);
    while (p_next_osm_port != (osm_port_t *)cl_qmap_end(&m_p_osm_subn->port_guid_tbl)) {
        p_osm_port = (osm_port_t *)p_next_osm_port;
        p_next_osm_port = (osm_port_t *)cl_qmap_next((const cl_map_item_t*)p_next_osm_port);

        CCNodeInfo node_info(
            cl_ntoh64(osm_port_get_guid(p_osm_port)),    // GUID
            cl_ntoh16(osm_port_get_base_lid(p_osm_port)),// LID
            p_osm_port->p_node,                          // osm_node_t
            p_osm_port->p_physp,                         // osm_physp_t
            m_osm_event_id == OSM_EVENT_ID_SUBNET_UP ? GetSL(p_osm_port) : m_p_osm_subn->opt.sm_sl);// SL

        switch (osm_node_get_type(p_osm_port->p_node)) {
            case IB_NODE_TYPE_CA:
                rc = UpdateCANode(node_info);
                break;
            case IB_NODE_TYPE_SWITCH:
                rc = UpdateSWNode(node_info);
                break;
            case IB_NODE_TYPE_ROUTER:
                break;
            default:
                rc = 1;
                break;
        }
    }
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_DEBUG,
        "Number of hosts for calculations is %d\n", m_num_hosts);
    return rc;
}


//============================================================================//
int CongestionControlManager::Run()
{
    int rc = 0;
    OSM_CC_LOG(m_p_osm_log, OSM_LOG_INFO, "Started configuration cycle\n");

    static unsigned cycle_num = 0;
    ++cycle_num;
    try {
        RemoveAbsentNodes();

        // Parse File
        ParseInputFile();
        UpdateUserOptions();
        UpdateFabricNodes();

        //Initialize - if already been initialized than do nothing
        rc = Init();

        // Collect Information on Nodes ( cc support )
        CollectCCSupportForAllNodes();

        // Set Congestion Parameters for all nodes
        SetCCParamsForAllNodes();
        m_cc_config_up = true;

        // On successful execution create CC Statistics Agent
        CreateCCStatisticsAgent();

        ReportFit(FIT_REPORT_CC_MGR_CYCLE_DONE, 1, &rc);
    }
    catch (int error) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_ERROR,
                "Caught exception = %u, exiting CC Manager...\n", error);
        rc = 1;
        ReportFit(FIT_REPORT_CC_MGR_CYCLE_DONE, 1, &rc);
    }
    m_cc_mgr_request = false;


    if (rc) {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_INFO, "Configuration cycle (number %u) failed\n", cycle_num);
        fprintf(stdout, "CC Manager - Configuration cycle (number %u) failed\n", cycle_num);
    } else {
        OSM_CC_LOG(m_p_osm_log, OSM_LOG_INFO, "Configuration cycle (number %u) completed successfully\n", cycle_num);
        fprintf(stdout, "CC Manager - Configuration cycle (number %u) completed successfully\n", cycle_num);
    }

    UnLockCCMutex(LOCK_HELD_BY_MGR);
    return rc;
}

